/**
 * Relbit flock
 * Replaces standard flock() and fclose() functions with it's own implementation
 * so the locks work properly on NFS.
 * 
 * Developed by Jakub Skokan <aither@havefun.cz> for Relbit LLC
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "php.h"
#include "php_ini.h"
#include "php_relbit_flock.h"
#include "ext/standard/flock_compat.h"
#include "ext/hash/php_hash.h"
#include "ext/hash/php_hash_sha.h"

#include <unistd.h>

ZEND_DECLARE_MODULE_GLOBALS(relbit_flock)

ZEND_BEGIN_ARG_INFO_EX(arginfo_flock, 0, 0, 2)
	ZEND_ARG_INFO(0, fp)
	ZEND_ARG_INFO(0, operation)
	ZEND_ARG_INFO(1, wouldblock)
ZEND_END_ARG_INFO()

static function_entry relbit_flock_functions[] = {
	PHP_FE(relbit_flock, arginfo_flock)
	PHP_FE(relbit_fclose, NULL)
	{NULL, NULL, NULL}
};

zend_module_entry relbit_flock_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
	STANDARD_MODULE_HEADER,
#endif
	PHP_RELBIT_FLOCK_EXTNAME,
	relbit_flock_functions,
	PHP_MINIT(relbit_flock),
	PHP_MSHUTDOWN(relbit_flock),
	PHP_RINIT(relbit_flock),
	NULL,
	PHP_MINFO(relbit_flock),
#if ZEND_MODULE_API_NO >= 20010901
	PHP_RELBIT_FLOCK_VERSION,
#endif
	STANDARD_MODULE_PROPERTIES
};

#define RELBIT_DEFAULT_LOCK_DIR "/mnt/.locks"
#define RELBIT_DEFAULT_TIMEOUT 15

PHP_INI_BEGIN()
	STD_PHP_INI_ENTRY("relbit_flock.lock_dir", RELBIT_DEFAULT_LOCK_DIR, PHP_INI_SYSTEM, OnUpdateString, lock_dir, zend_relbit_flock_globals, relbit_flock_globals)
	STD_PHP_INI_ENTRY("relbit_flock.lock_timeout", "15", PHP_INI_ALL, OnUpdateLong, timeout, zend_relbit_flock_globals, relbit_flock_globals)
PHP_INI_END()

#ifdef COMPILE_DL_RELBIT_FLOCK
ZEND_GET_MODULE(relbit_flock)
#endif

/* These constants are copied from flock_compat.h, since there's some stupid ifdef
 * and they just won't define
 */
#define LOCK_SH 1
#define LOCK_EX 2
#define LOCK_NB 4
#define LOCK_UN 8

static int flock_values[] = { LOCK_SH, LOCK_EX, LOCK_UN };

char* php_relbit_get_sha256(char *str)
{
	void *context;
	char *data, *digest, *hex_digest;
	
	// These two values are extracted from ext/hash/hash_sha.c, since I cannot use hash abstract interface
	// They are valid only for sha256
	int context_size = sizeof(PHP_SHA256_CTX), digest_size = 32;
	
	context = emalloc(context_size);
	PHP_SHA256Init(context);
	
	PHP_SHA256Update(context, (unsigned char *) str, strlen(str));
	
	digest = emalloc(digest_size + 1);
	PHP_SHA256Final((unsigned char *) digest, context);
	efree(context);
	
	hex_digest = safe_emalloc(digest_size, 2, 1);

	php_hash_bin2hex(hex_digest, (unsigned char *) digest, digest_size);
	hex_digest[2 * digest_size] = 0;
	efree(digest);
	
	return hex_digest;
}

int php_relbit_replace_function(char *what, char *with)
{
	zend_function *func;
	
	if(zend_hash_find(CG(function_table), with, strlen(with)+1, (void **) &func) == FAILURE)
	{
		zend_error(E_ERROR, "%s() function name not present in global function_table", with);
		return FAILURE;
	}
	
	// Not sure what this does, but, well, it won't harm us
	function_add_ref(func);
	
	zend_hash_del(CG(function_table), what, strlen(what)+1);
	
	if(zend_hash_add(CG(function_table), what, strlen(what)+1, func, sizeof(zend_function), NULL) == FAILURE)
	{
		zend_error(E_ERROR, "failed to put %s() to function_table", with);
		return FAILURE;
	}
	
	zend_hash_del(CG(function_table), with, strlen(with)+1);
}

int php_relbit_get_link_target(long fd, char *target, int size)
{
	int ret;
	char link[30];
	
	// Yes, it always returns fd + 2, who knows why...
	sprintf(link, "/proc/self/fd/%ld", fd - 2);
	
	ret = readlink(link, target, size-1);
	
	if(ret == -1)
	{
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "%s", strerror(errno));
		return 0;
	}
	/* Append NULL to the end of the string */
	target[ret] = '\0';
	
	return 1;
}

int php_relbit_unlock(char *target, char *path, char *hex_digest)
{
	FILE *lock_file;
	zval **value;
	
	//printf("Unlocking... %s\n", hex_digest);
	
	if(zend_hash_find(Z_ARRVAL_P( RELBIT_FLOCK_G(locks) ), hex_digest, strlen(hex_digest)+1, (void **) &value) == FAILURE)
	{
		//zend_error(E_ERROR, "%s() function name not present in global function_table", with);
		//printf("hash '%s' not found - this file was not locked by us, let it be\n", hex_digest);
		
		return 1;
	}
	
	zend_hash_del(Z_ARRVAL_P( RELBIT_FLOCK_G(locks) ), hex_digest, strlen(hex_digest)+1);
	
	if( (lock_file = fopen(path, "r")) != NULL )
	{
		fclose(lock_file);
	
		if(unlink(path) == -1)
		{
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Cannot unlock file %s: cannot remove lock file %s", target, path);
			return 0;
		}
	}
	return 1;
}

PHP_FUNCTION(relbit_flock)
{
	zval *file;
	long operation;
	zval *wouldblock = NULL;
	php_stream *stream;
	struct stat sb;
	int act;
	char target[MAXPATHLEN];
	char path[MAXPATHLEN];
	int ret, pos = 0;
	FILE *lock_file;
	char *hex_digest;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl|z", &file, &operation, &wouldblock) == FAILURE)
	{
		RETURN_FALSE;
	}
	
	act = operation & 3;
	if (act < 1 || act > 3)
	{
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Illegal operation argument");
		RETURN_FALSE;
	}
	
	act = flock_values[act - 1] | (operation & PHP_LOCK_NB ? LOCK_NB : 0);
	
	// This will validate given resource, otherwise we don't need it
	php_stream_from_zval(stream, &file);
	
	if (stream == NULL)
	{
		RETURN_FALSE;
	}
	
	if(wouldblock && PZVAL_IS_REF(wouldblock)) {
		convert_to_long_ex(&wouldblock);
		Z_LVAL_P(wouldblock) = 0;
	}

	if(!php_relbit_get_link_target(Z_RESVAL_P(file), target, MAXPATHLEN))
	{
		RETURN_FALSE;
	}
	
	//printf("Link resolved to: %s\n", target);
	
	hex_digest = php_relbit_get_sha256(target);
	
	strcpy(path, RELBIT_FLOCK_G(lock_dir));
	strcat(path, "/");
	strcat(path, hex_digest);
	
	if(act & LOCK_UN)
	{
		if(php_relbit_unlock(target, path, hex_digest))
		{
			RETURN_TRUE;
		} else {
			RETURN_FALSE;
		}
	}
	
	// We'll wait till file is unlocked
	//printf("Waiting till file is unlocked... %s\n", path);
	
	do {
		if((lock_file = fopen(path, "r")) == NULL)
			break;
		fclose(lock_file);
		
		if(!stat(path, &sb) && time(NULL) > (sb.st_mtime + RELBIT_FLOCK_G(timeout)))
		{
			lock_file = NULL;
			break;
		}
		
		if(act & LOCK_NB)
			break;
		
		usleep(10000);
	} while(lock_file != NULL);
	
	if(lock_file != NULL)
	{
		if(wouldblock && PZVAL_IS_REF(wouldblock))
			Z_LVAL_P(wouldblock) = 1;
		
		RETURN_FALSE;
	}
	
	// Now lock it for us
	//printf("Locking...\n");
	if( (lock_file = fopen(path, "w")) == NULL )
	{
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Cannot flock file %s: cannot create lock file %s", target, path);
		RETURN_FALSE;
	}
	fclose(lock_file);
	
	add_assoc_null(RELBIT_FLOCK_G(locks), hex_digest);
	
	efree(hex_digest);
	
	RETURN_TRUE;
}

PHP_FUNCTION(relbit_fclose)
{
	zval *arg1;
	php_stream *stream;
	char target[MAXPATHLEN];
	char path[MAXPATHLEN];
	char *hex_digest;
	int ret;
	FILE *lock_file;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &arg1) == FAILURE) {
		RETURN_FALSE;
	}
	
	php_stream_from_zval_no_verify(stream, &arg1);
	if(stream == NULL)
	{
		RETURN_FALSE;
	}

	if ((stream->flags & PHP_STREAM_FLAG_NO_FCLOSE) != 0) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "%d is not a valid stream resource", stream->rsrc_id);
		RETURN_FALSE;
	}

	if(!php_relbit_get_link_target(Z_RESVAL_P(arg1), target, MAXPATHLEN))
	{
		RETURN_FALSE;
	}
	
	hex_digest = php_relbit_get_sha256(target);
	
	strcpy(path, RELBIT_FLOCK_G(lock_dir));
	strcat(path, "/");
	strcat(path, hex_digest);
	
	if (!stream->is_persistent) {
		php_stream_close(stream);
	} else {
		php_stream_pclose(stream);
	}
	
	ret = php_relbit_unlock(target, path, hex_digest);
	
	efree(hex_digest);
	
	if(ret)
	{
		RETURN_TRUE;
	} else {
		RETURN_FALSE;
	}
}

static void php_relbit_flock_init_globals(zend_relbit_flock_globals *relbit_flock_globals)
{
	relbit_flock_globals->lock_dir = RELBIT_DEFAULT_LOCK_DIR;
	relbit_flock_globals->timeout = RELBIT_DEFAULT_TIMEOUT;
}

PHP_MINIT_FUNCTION(relbit_flock)
{
	ZEND_INIT_MODULE_GLOBALS(relbit_flock, php_relbit_flock_init_globals, NULL);
	
	REGISTER_INI_ENTRIES();
	
	if(php_relbit_replace_function("flock", "relbit_flock") == SUCCESS && php_relbit_replace_function("fclose", "relbit_fclose") == SUCCESS)
		return SUCCESS;
	else return FAILURE;
}

PHP_RINIT_FUNCTION(relbit_flock)
{
	ALLOC_INIT_ZVAL(RELBIT_FLOCK_G(locks));
	array_init(RELBIT_FLOCK_G(locks));
}

PHP_MSHUTDOWN_FUNCTION(relbit_flock)
{
	UNREGISTER_INI_ENTRIES();
	
	return SUCCESS;
}

PHP_MINFO_FUNCTION(relbit_flock)
{
	char timeout[10];
	sprintf(timeout, "%d", RELBIT_FLOCK_G(timeout));
	
	php_info_print_table_start();
	php_info_print_table_header(2, "Relbit flock", "enabled");
	php_info_print_table_row(2, "Version", PHP_RELBIT_FLOCK_VERSION);
	php_info_print_table_row(2, "relbit_flock.lock_dir", RELBIT_FLOCK_G(lock_dir));
	php_info_print_table_row(2, "relbit_flock.lock_timeout", timeout);
	php_info_print_table_end();
}
